Location tanC[2], tanG[2];
int8_t getRegion(uint8_t onlyDistance, Location *current); 

void resetEnc(void)
{
	encDistance = 0;
	enc = encoder();
	encRunning = 1;
}

float getEncDistance(void)
{
	if(!encRunning) {
		enc = encoder();
	}
	
	encDistance += (encoder()-enc);
	enc = encoder();
	encRunning = 1;
	
	return encDistance*ENC_TO_VPS;
}

void pauseEnc(void)
{
	encRunning = 0;
}

int16_t limitVel(float vel) 
{
	if (vel < -255) return -255; 
	if (vel > 255) return 255;
	return (int16_t)vel;
}

void stopRobot()
{
	motorRight(0);
	motorLeft(0);
}
void setVelocities(float correctionAmount)
{
	motorRight(motorDirection*limitVel(ForwardVel + motorDirection*correctionAmount));
	motorLeft(motorDirection*limitVel(ForwardVel - motorDirection*correctionAmount));
}

void forwardRotate(float deg)
{
//	const float minSpeed = 240, factor = 10;
	
 	//motorRight(limitVel(signOf(deg)*max(minSpeed,fabs(deg)*factor)));
 	//motorLeft(limitVel(-signOf(deg)*max(minSpeed,fabs(deg)*factor)));
	
 	motorRight(limitVel(signOf(deg)*ROTATION_VEL));
 	motorLeft(limitVel(-signOf(deg)*ROTATION_VEL));
	
}

void rotate(float angle) {
	
	angle *= 0.96;
	stopRobot();
	pause(10);
	gyro_set_degrees(0);
	uint32_t lastTime = get_time();
	float lastAngle = 0;
	
	while (1) {  //TODO add "apoptosis" after a certain time (that is proportional to the angle) in case it got stuck
		if (fabs(gyro_get_degrees())>=fabs(angle)) break;
		
		if(get_time() - lastTime > 200) {
			if(fabs(lastAngle-gyro_get_degrees()) < 4.0) {
				stopRobot();
				motorRight(-motorDirection*220);
				motorLeft(-motorDirection*220);
				pause(180);
				stopRobot();
			}
			
			lastAngle = gyro_get_degrees();
			lastTime = get_time();
		}
		
		forwardRotate(angle);
	}
	stopRobot();
	pause(90);
}

void moveDistance(float distance) //don't forget to set motorDirection before using this!
{
	struct pid_controller headingPID;

  	init_pid(&headingPID,KP,  KI, KD, &gyro_get_degrees, &setVelocities);

	headingPID.goal = gyro_get_degrees();
  	headingPID.enabled = true;
  	
  	resetEnc();
  	
  	while(getEncDistance() < distance - 150)
  		update_pid(&headingPID);
  	
  	stopRobot();
}

void moveTillCollide(int16_t speed)
{
	float dis = -1;
	
	struct pid_controller headingPID;

  	init_pid(&headingPID,KP,  KI, KD, &gyro_get_degrees, &setVelocities);

	resetEnc();
	ForwardVel = speed;
	motorRight(speed);
	motorLeft(speed);
	
	while(1) {
		pause(120);
		if(dis == getEncDistance()) {
			stopRobot();
			ForwardVel = 160;
			return;
		}
		
		dis = getEncDistance();
	}
}	

int checkBarrier(Location *current, Location *goal)
{	
	float dis = getDistance(current, goal);
	float r = (goal->x - current->x)*(float)(-goal->y) - (goal->y - current->y)*(float)(-goal->x);
	
	if(dis < 1) return 0;
	r = fabs(r/dis);
	
	if(r > CIRCLE_RADIUS) return 0; //good to go = all clear
	else return 1; //go through an alternative path cuz there's a barrier -> gotta get the hell outta there :)

}
void setVelocity(int16_t distance){

	if(isExploring) {ForwardVel = 240; return;}
		
	uint16_t DistLowerLimit = 500, DistUpperLimit = 900,VelLowerLimit = 100, VelUpperLimit = 235;

	if(goingTo == 2) VelLowerLimit = 100; //mines
	else if(goingTo == 1) VelLowerLimit = 100; //gearbox
	
	if (distance<= DistLowerLimit) {
		ForwardVel= VelLowerLimit;
	} else if (distance>= DistUpperLimit) {
		ForwardVel= VelUpperLimit;
	} else {
		float slope=(float)(VelUpperLimit-VelLowerLimit)/(DistUpperLimit-DistLowerLimit);
		ForwardVel= slope*distance+(VelUpperLimit-slope*DistUpperLimit);
	}
}
int8_t getLineCirleIntersection(Location *point, Location *reta, Location *retb)
{
	//This is a check for safety.
	if((float)point->x*point->x + (float)point->y*point->y < (float)CIRCLE_RADIUS*CIRCLE_RADIUS)
		return 0;
		
	float a,b,c,s,y1,y2;
	
	if(!point->x) point->x = 1;
	if(!point->y) point->y = 1;
	
	a = (1.0+(float)(point->y)*(point->y)/((float)point->x*point->x));
	b = -2.0*point->y*CIRCLE_RADIUS*CIRCLE_RADIUS/((float)point->x*point->x);
	c = ((float)CIRCLE_RADIUS*CIRCLE_RADIUS*CIRCLE_RADIUS*CIRCLE_RADIUS)/((float)point->x*point->x) - (float)CIRCLE_RADIUS*CIRCLE_RADIUS;
		
	s = b*b - 4*a*c;
	if(s<0.0) return 0;
	
	y1 = (-b + sqrt(s))/(2*a);
	y2 = (-b - sqrt(s))/(2*a);
	
	reta->y = getFarther(y1); retb->y = getFarther(y2);
	
	reta->x = getFarther(((float)CIRCLE_RADIUS*CIRCLE_RADIUS - y1*point->y)/point->x);
	retb->x = getFarther(((float)CIRCLE_RADIUS*CIRCLE_RADIUS - y2*point->y)/point->x);
	
	return 1;
}


// return the intersection point between the two tangent lines 
Location alternativePoint(Location *current, Location *goal) {
	if(!getLineCirleIntersection(current, &tanC[0], &tanC[1]) ||
		!getLineCirleIntersection(goal, &tanG[0], &tanG[1]))
		{
			printf("Panic at alternativePoint(): getLineCirleIntersection() returned zero!");
			return *goal;
		}

	int8_t i, j, locR, oppR, curR, goalR;
	float t, dis, best = 10e10;
	Location bestLoc, loc, opponent = getOpponentLocation();
	
	curR = getRegion(1, current);
	goalR = getRegion(1, goal);
	oppR = getRegion(1, &opponent);
	
	bestLoc = midpoint[(curR+goalR)/2]; // TODO should add something to handle the case where abs(curR-curGoal) = 3
	
	for(i=0;i<2;++i) 
		for(j=0;j<2;++j) {
			t = (float)(-tanC[i].y)*tanG[j].x - (float)(tanC[i].x)*(-tanG[j].y);
			
			if(fabs(t) < 0.01) continue; //t ~ 0 (division by zero to come); lines are parallel
			t = ((float)(-tanG[j].y)*(tanC[i].y - tanG[j].y) - (float)(tanG[j].x)*(tanC[i].x - tanG[j].x))/t;
			
			loc.x = getFarther(tanC[i].x + (-tanC[i].y)*t);
			loc.y = getFarther(tanC[i].y + tanC[i].x*t);
			
			loc.x = signOf(loc.x)*min(1400, abs(loc.x));
			loc.y = signOf(loc.y)*min(1400, abs(loc.y));
			
			//if(abs(loc.x) > 1800 || abs(loc.y) > 1800) continue;
			
			//account for other robot!
			locR = getRegion(1, &loc);
			if(curR > goalR) {
				if((locR > curR || locR < goalR) && (oppR > curR || oppR < goalR)) continue;
				if((locR < curR && locR > goalR) && (oppR < curR && oppR > goalR)) continue;
			}
			else {
				if((locR > goalR || locR < curR) && (oppR > goalR || oppR < curR)) continue;
				if((locR < goalR && locR > curR) && (oppR < goalR && oppR > curR)) continue;
			}
			
			dis = getDistance(current, &loc) + getDistance(&loc, goal);
			
			if(dis < best) {
				best = dis;
				bestLoc = loc;
			}
		}

	return bestLoc;
}

int8_t gotoLocation(int8_t *status, Location *goal, int8_t dir)
{
	float dis=-1;

	uint32_t lastTime = 0, time, encTime = get_time();
	float angle_Correction = 0, dist, angle, theta;
	Location diff; int8_t i, arrived = 0;
	int8_t goal_region = getRegion(1, goal);
	
	struct pid_controller headingPID;

  	init_pid(&headingPID, (isExploring)?8:KP,  KI, KD, &gyro_get_degrees, &setVelocities);

	gyro_set_degrees(0);
	headingPID.goal = 0;
  	headingPID.enabled = true;

	*status = 1;
	
	for(i=0;i<3;++i) {
	
		while(*status) {
			if(isExploring && get_time() - start_time > 9000) {
				isExploring = 0;
				stopRobot();
				return 4; //end exploration
			}
			
			diff = getVPSCurrentLocation(&theta, &time); // diff is the current location
			if(time==0 || time == lastTime) goto PID;
			lastTime = time;

			dist = getDistance(&diff,goal);
			diff.x = goal->x - diff.x; diff.y = goal->y - diff.y; // differance in location
			angle = atan2(diff.y, diff.x)*180.0/PI;
			if(dist<((isExploring)?1.5*LOCATION_TOLERANCE:((goingTo == 3)?330:LOCATION_TOLERANCE))) {
					if(!isExploring) stopRobot();
					arrived = 1;
					break;
			} 
		
			setVelocity(dist);
			
			//ForwardVel=180;
			 // correct theta if it is going backwards 
			if(motorDirection == -1) theta = deg_mod(theta+180.0);

			angle_Correction = getAngle(theta, angle);
		
			 // go backwards if angle_Correction is >90
			if((!dir && fabs(angle_Correction)>90.0) || (dir && dir != motorDirection)) {
				angle_Correction = -signOf(angle_Correction)*(180.0-fabs(angle_Correction));
				motorDirection *= -1;
			} 
	
PID:
			if(!isExploring && fabs(angle_Correction)>30) {
				rotate(angle_Correction);
				angle_Correction=0;
				pause(100); // TODO or skip the next VPS (2~4) readings 
				dis = -1;
				encTime = get_time();
				continue;
			}
			
			
			if(get_time() - encTime > 240) {
				if(dis == getEncDistance()) {
					if(getRegion(1, NULL) == goal_region && assumeArrival) {
						stopRobot();
						return 2;
					}
					else {
						motorRight(-motorDirection*250);
						motorLeft(-motorDirection*250);
						pause(250);
						stopRobot();
						
						if(getRegion(1, NULL) != goal_region) {
							if(isExploring) return 3; //3, for exploration, means "I'm stuck, use Smart"
							
							// TODO might end up in a stack overflow! Don't use recursion!
							//gotoLocation(&alwaysOne, &midpoint[getRegion(1, NULL)], 0);
							//will go to midpoint and try again...
						}
					}
				}
				
				dis = getEncDistance();
				encTime = get_time();
			}
			if(isExploring && fabs(angle_Correction)>20) ForwardVel=180;
			gyro_set_degrees(-angle_Correction*((ForwardVel>130)?0.75:.65));

			update_pid(&headingPID);
		}
		
		//ForwardVel = 160;
	//	if(arrived) return 1;
		return 1;
		//needs to make sure *status won't be overwritten back to zero before it actually moves.
	//	gotoLocation(status, &midpoint[getRegion(1, NULL)], 0);
		//return 0;
	}
	
	return 0;
}

void gotoLocationSmart(int8_t *status, Location *goal, int8_t dir)
{
	//TODO if it's very close to the inner hexagon, it might try to rotate, but it's stuck... fix that
	Location current = getVPSCurrentLocation(NULL, NULL);
	Location origin = {.x=0, .y=0};
	int8_t region;

	if(getDistance(&origin, &current) < CIRCLE_RADIUS) {
		region = getRegion(1, NULL);
		assumeArrival = 0;
		gotoLocation(status, &midpoint[region], 0);
		assumeArrival = 1;
		current = getVPSCurrentLocation(NULL, NULL);
	}

	if(getDistance(&origin, goal) < CIRCLE_RADIUS) {
		region = getRegion(1, goal);
	
		if(checkBarrier(&current, &midpoint[region])) {
			origin = (alternativePoint(&current, &midpoint[region]));
			gotoLocation(status, &origin, 0);
		}
		gotoLocation(status, &midpoint[region], 0);
		assumeArrival = 1;
		gotoLocation(status, goal, dir);
	}
	else {
		if(checkBarrier(&current, goal)) {
			Location alt = alternativePoint(&current, goal);
			gotoLocation(status, &alt, 0);
		}
		assumeArrival = 1;
		gotoLocation(status, goal, dir);
	}
}
